"""
Base of taped objects

This module provides classes and functions for handling audio streams and buffers.
It includes tools for capturing live audio from microphones, buffering audio data,
and converting between different audio representations.

Examples
--------

"""

from itertools import chain
from typing import Union, Optional
from collections import namedtuple
from dataclasses import dataclass

from stream2py.stream_buffer import StreamBuffer
from audiostream2py import PyAudioSourceReader, AudioSegment
from taped.util import (
    DFLT_SR,
    DFLT_SAMPLE_WIDTH,
    DFLT_CHK_SIZE,
    DFLT_STREAM_BUF_SIZE_S,
    bytes_to_waveform,
    ensure_source_input_device_index,
)

from itertools import islice
from creek import Creek

BufferItemOutput = namedtuple(
    typename="BufferItemOutput",
    field_names=["timestamp", "bytes", "frame_count", "time_info", "status_flags"],
)


def audio_segment_to_buffer_item_output(segment: AudioSegment) -> BufferItemOutput:
    return BufferItemOutput(
        timestamp=segment.start_date,
        bytes=segment.waveform,
        frame_count=segment.frame_count,
        time_info=segment.end_date,
        status_flags=segment.status_flags,
    )


@Creek.wrap
@dataclass
class BaseBufferItems(StreamBuffer):
    """A generator of live chunks of audio bytes taken from a stream sourced from specified microphone.

    :param input_device_index: Index of Input Device to use. Unspecified (or None) uses default device.
    :param sr: Specifies the desired sample rate (in Hz)
    :param sample_bytes: Sample width in bytes (1, 2, 3, or 4)
    :param sample_width: Specifies the number of frames per buffer.
    :param stream_buffer_size_s: How many seconds of data to keep in the buffer (i.e. how far in the past you can see)
    """

    input_device_index: Optional[Union[int, str]] = None
    sr: int = DFLT_SR
    sample_width: int = DFLT_SAMPLE_WIDTH
    chk_size: int = DFLT_CHK_SIZE
    stream_buffer_size_s: Union[float, int] = DFLT_STREAM_BUF_SIZE_S
    verbose: bool = False

    def __post_init__(self):
        self.input_device_index = ensure_source_input_device_index(
            self.input_device_index, verbose=self.verbose
        )
        seconds_per_read = self.chk_size / self.sr

        self.maxlen = int(self.stream_buffer_size_s / seconds_per_read)
        self.source_reader = PyAudioSourceReader(
            rate=self.sr,
            width=self.sample_width,
            unsigned=True,
            input_device_index=self.input_device_index,
            frames_per_buffer=self.chk_size,
        )

        super().__init__(source_reader=self.source_reader, maxlen=self.maxlen)


class BufferItems(BaseBufferItems):
    def data_to_obj(self, data):
        return audio_segment_to_buffer_item_output(data)


class ByteChunks(BufferItems):
    def data_to_obj(self, data):
        return super().data_to_obj(data).bytes


# TODO: use more stable and flexible bytes_to_waveform
class WfChunks(ByteChunks):
    sample_width = 2

    def data_to_obj(self, data):
        data = super().data_to_obj(data)
        return bytes_to_waveform(data, sample_width=self.sample_width)


# TODO: use more stable and flexible bytes_to_waveform
class LiveWf(WfChunks):
    def post_iter(self, obj):
        return chain.from_iterable(obj)

    @property
    def available_samples_in_buffer(self):
        """Returns the actual max number of samples available.
        This is maxlen * chk_size when buffer is full.
        """

        return len(self.source_buffer._buffer) * self.chk_size

    def __getitem__(self, item):
        if not isinstance(item, slice):
            item = slice(item, item + 1)  # to emulate usual list[i] interface
        item = positive_slice_version(item, self.available_samples_in_buffer)
        return list(islice(self, item.start, item.stop, item.step))


def positive_slice_version(slice_, sliced_obj_len):
    """Returns a slice where negative start and stops are replaced with their positive correspondence.
    This is because `itertools.islice` doesn't accept negative entries, so we need to tell it
    explicitly what we mean by -3 (we mean sliced_obj_len - 3).

    >>> positive_slice_version(slice(-7, -1, 3), 10)
    slice(3, 9, 3)
    >>> positive_slice_version(slice(2, -2), 7)
    slice(2, 5, None)
    """

    def positivize(x):
        if x is not None and x < 0:
            x += sliced_obj_len
        return x
        # return x if x is None or x > 0 else 0

    return slice(positivize(slice_.start), positivize(slice_.stop), slice_.step)



from itertools import islice
from functools import partial
from statistics import pstdev
import pickle
import os

from i2.signatures import Sig
from lined import Pipeline, iterize, iterate

from taped.util import wraps
from taped import LiveWf, simple_chunker


def std(x):
    """Wrapping statistics.pstdev, so it doesn't cause issues such as https://bugs.python.org/issue39218"""
    return pstdev(map(float, x))


def mk_pipeline(
    chunker=partial(simple_chunker, chk_size=2048), model=pstdev, output_func=print
):
    if isinstance(chunker, int):
        chk_size = chunker
        chunker = partial(simple_chunker, chk_size=chk_size)
    return Pipeline(chunker, iterize(model), iterize(output_func), iterate)


def launch(pipeline=None, max_samples=20000):
    if pipeline is None:
        pipeline = mk_pipeline()
    try:
        with LiveWf() as live_wf:
            if max_samples:
                wf = islice(live_wf, 0, max_samples)
            else:
                wf = live_wf
            pipeline(wf)
    except KeyboardInterrupt:
        print("A KeyboardInterrupt was received. Closing down...")


def viz(val, gain=1 / 20, offset=0, disp_str="*"):
    print(disp_str * int(gain * val + offset))


def _unpickle_if_filepath(obj):
    if isinstance(obj, str) and os.path.isfile(obj):
        filepath = obj
        with open(filepath, "r") as f:
            obj = pickle.load(f)
    return obj


def strings_as_pickle_files(func):
    @wraps(func)
    def _func(*args, **kwargs):
        _args = tuple(map(_unpickle_if_filepath, args))
        _kwargs = {k: _unpickle_if_filepath(v) for k, v in kwargs.items()}
        return func(*_args, **_kwargs)

    return _func


# Pattern: TODO: Use general input_trans pattern
@strings_as_pickle_files
@wraps(mk_pipeline)
def main(*args, **kwargs):
    max_samples = kwargs.pop(
        "max_samples", None
    )  # this argument will be hidden (use Sig to add it to signature?)
    kwargs = Sig(mk_pipeline).extract_kwargs(*args, **kwargs)
    if "chunker" in kwargs and str.isnumeric(kwargs["chunker"]):
        kwargs["chunker"] = int(kwargs["chunker"])
    pipeline = mk_pipeline(**kwargs)

    return launch(pipeline, max_samples=max_samples)


if __name__ == "__main__":
    import argh
    from inspect import signature

    argh.dispatch_command(main)
    # argh.dispatch_command(main, argv=list(signature(mk_pipeline).parameters))



"""This is a first draft to get some functions to estimate sample rate
bias and drift.

Consider that functions might have to be written that are closer to the microphone to
get more accurate estimates.

"""

import os
from itertools import islice
from functools import partial
from typing import Sequence, MutableMapping, Union
from time import time
import pickle
import json
from pathlib import Path

import taped
from lkj import get_watermarked_dir
from slang import fixed_step_chunker
from dol import Files, wrap_kvs, Pipe

# import pandas as pd


def gather_sample_rate_date(
    chk_size=20000,
    sr=20_000,
    n_chks=10,
    input_device_index=None,
    read_chk_size=4096,
    *,
    display_input_device_info_when_not_chosen=False,
):
    """"""
    chunker = partial(fixed_step_chunker, chk_size=chk_size)
    wf = taped.LiveWf(
        input_device_index=input_device_index, sr=sr, chk_size=read_chk_size
    )
    with wf:

        def gen():
            chks = iter(islice(chunker(wf), n_chks))
            while True:
                try:
                    yield {"bt": time(), "n_samples": len(next(chks)), "tt": time()}
                except StopIteration:
                    break

        data = list(gen())

    return data


def params_to_key(params: dict, ext="") -> str:
    """Convert a dict of parameters to a key for a dict-like object

    >>> params_to_key({'a': 1, 'b': 2})
    'a=1,b=2'
    >>> params_to_key({'b': 2, 'a': 1}, ext='.csv')
    'a=1,b=2.csv'
    """
    return ",".join(f"{k}={params[k]}" for k in sorted(params)) + ext


def key_to_params(key: str):
    """Convert a key to a dict of parameters. Approximate inverse of params_to_key

    >>> key_to_params('a=1,b=2')
    {'a': '1', 'b': '2'}
    >>> key_to_params('b=2,a=1.csv')
    {'a': '1', 'b': '2'}
    """
    key, _ = os.path.splitext(key)  # remove the extension if present
    kvs = (kv.split("=") for kv in key.split(","))
    return dict(kv for kv in sorted(kvs, key=lambda x: x[0]))


dflt_store_folder = get_watermarked_dir(
    "taped/sample_rate_bias", make_dir=partial(os.makedirs, exist_ok=True)
)
dflt_params_folder = os.path.join(dflt_store_folder, "params")
JsonFiles = wrap_kvs(
    Files, data_of_obj=Pipe(json.dumps, str.encode), obj_of_data=json.loads
)
dflt_store = JsonFiles(dflt_store_folder, max_levels=0)

import itertools

list(itertools.product([1, 2], [4, 5, 6]))

IntList = Sequence[int]


def load_json(filepath: str) -> dict:
    return json.loads(Path(filepath).read_text())


def save_json(filepath: str, data: Union[dict, list]):
    Path(filepath).write_text(json.dumps(data))


def params_product(
    save_to_filepath: str = None,
    *,
    chk_size: IntList = (20_000,),
    sr: IntList = (20_000,),
    n_chks: IntList = (30,),
    read_chk_size: IntList = (4096,),
):
    """To create a list of parameters to pass to run_experiments"""
    params = [
        dict(zip(("chk_size", "sr", "n_chks", "read_chk_size"), p))
        for p in itertools.product(chk_size, sr, n_chks, read_chk_size)
    ]
    if save_to_filepath:
        save_json(save_to_filepath, params)


def _get_params_list(params_list: Union[str, Sequence[dict]]) -> Sequence[dict]:
    if isinstance(params_list, str):
        import json
        from pathlib import Path

        filepath = params_list
        if not os.path.exists(filepath):
            filepath = os.path.join(dflt_params_folder, filepath)

        params_list = json.loads(Path(filepath).read_text())
    return params_list


def _get_store(store: Union[str, MutableMapping] = dflt_store) -> MutableMapping:
    if isinstance(store, str):
        filepath = store
        store = JsonFiles(filepath)
    return store


def run_experiments(
    params_list: Union[str, Sequence[dict]],
    store: Union[str, MutableMapping] = dflt_store,
    *,
    print_progress=True,
    overwrite: bool = False,
):
    params_list = _get_params_list(params_list)
    store = _get_store(store)
    n = len(list(params_list))
    for i, params in enumerate(params_list, 1):
        key = params_to_key(params)
        if key in store and not overwrite:
            continue  # skip this one, we have it already
        print_progress and print(f"{i}/{n}: {key}")
        store[key] = gather_sample_rate_date(**params)
    return store


def observed_sample_rates(d):
    import pandas as pd

    df = pd.DataFrame(d)
    return df["n_samples"] / (df["tt"] - df["bt"])


def observed_sample_rates_plot(k, v):
    import pandas as pd

    df = pd.DataFrame(v)
    df = df["n_samples"] / (df["tt"] - df["bt"])
    return df.plot(title=k, marker="o", linestyle="-")


def mk_df_store(store):
    return wrap_kvs(store, obj_of_data=observed_sample_rates)


def mk_plot_store(store):
    return wrap_kvs(store, postget=observed_sample_rates_plot)


if __name__ == "__main__":
    import argh

    argh.dispatch_command(run_experiments)



"""
Recording and playback tools for audio data.
"""

from typing import Callable, Iterable, List, Literal, Optional, Tuple, Union
from itertools import islice
import soundfile as sf
from taped.base import BaseBufferItems, LiveWf
from taped.util import DFLT_SR, DFLT_SAMPLE_WIDTH, DFLT_CHK_SIZE, DFLT_STREAM_BUF_SIZE_S


def record(
    duration: Optional[float] = None,
    *,
    duration_unit: Literal["seconds", "samples", "minutes"] = "seconds",
    sr: int = DFLT_SR,
    egress: Optional[Union[str, Callable]] = None,
    ignore_exceptions: Tuple[Exception, ...] = (KeyboardInterrupt,),
    input_device_index: Optional[int] = None,
    sample_width: int = DFLT_SAMPLE_WIDTH,
    chk_size: int = DFLT_CHK_SIZE,
    stream_buffer_size_s: float = DFLT_STREAM_BUF_SIZE_S,
    verbose: bool = False,
) -> List:
    """Record audio and return waveform data.

    Args:
        duration: Length of recording. If None, records until interrupted.
        duration_unit: Unit for duration ('seconds', 'samples', or 'minutes').
        sr: Sample rate.
        egress: Function to process waveform before returning or filename to save to.
        ignore_exceptions: Exceptions to catch and exit cleanly.
        input_device_index: Index of input device to use.
        sample_width: Sample width in bytes.
        chk_size: Chunk size for reading audio.
        stream_buffer_size_s: Buffer size in seconds.
        verbose: Whether to print status messages.

    Returns:
        Recorded waveform data, potentially processed by egress function.

    >>> # Record 0.1 seconds of audio
    >>> sample = record(0.1, verbose=False)  # doctest: +SKIP
    """

    def _log(*args, **kwargs):
        if verbose:
            print(*args, **kwargs)

    def _save_to_file(waveform, filename):
        """Save waveform to file and return the waveform."""
        sf.write(filename, waveform, samplerate=sr)
        return waveform

    def _convert_duration(value, unit):
        """Convert duration from given unit to number of samples."""
        if value is None:
            return None

        if unit == "seconds":
            return int(value * sr)
        elif unit == "minutes":
            return int(value * 60 * sr)
        elif unit == "samples":
            return int(value)
        else:
            raise ValueError(f"Unknown duration unit: {unit}")

    # Convert duration to samples
    n_samples = _convert_duration(duration, duration_unit)

    # Determine egress function
    if egress is None:
        _egress = lambda wf: wf
    elif isinstance(egress, str):
        _egress = lambda wf: _save_to_file(wf, egress)
    else:
        _egress = egress

    # Initialize empty waveform before trying anything
    waveform = []

    try:
        if input_device_index is None and sample_width == DFLT_SAMPLE_WIDTH:
            # Use the simpler LiveWf approach when possible
            _log("Starting recording with LiveWf...")
            with LiveWf(sr=sr) as live_wf:
                # Accumulate samples but be prepared for interruption
                if n_samples is not None:
                    live_iter = islice(live_wf, n_samples)
                else:
                    live_iter = live_wf

                # Collect samples one by one to ensure we keep what we have on interrupt
                for sample in live_iter:
                    waveform.append(sample)
        else:
            # Use the more configurable BaseBufferItems approach
            _log("Starting recording with BaseBufferItems...")
            buffer_items = BaseBufferItems(
                input_device_index=input_device_index,
                sr=sr,
                sample_width=sample_width,
                chk_size=chk_size,
                stream_buffer_size_s=stream_buffer_size_s,
            )

            count = 0

            with buffer_items:
                _log("Recording started (interrupt to stop)...")
                for item in buffer_items:
                    waveform.extend(item.data)
                    count += len(item.data)

                    if n_samples is not None and count >= n_samples:
                        # Trim to exact length if needed
                        waveform = waveform[:n_samples]
                        break

    except ignore_exceptions as e:
        _log(f"Recording stopped by {type(e).__name__}")
    except Exception as e:
        _log(f"Error during recording: {type(e).__name__}: {e}")
        raise

    _log(f"Recorded {len(waveform)} samples")
    return _egress(waveform)


# TODO: Deprecate record_some_sound
def record_some_sound(
    save_to_file,
    input_device_index=None,
    sr=DFLT_SR,
    sample_width=DFLT_SAMPLE_WIDTH,
    chk_size=DFLT_CHK_SIZE,
    stream_buffer_size_s=DFLT_STREAM_BUF_SIZE_S,
    verbose=True,
):
    def get_write_file_stream():
        if isinstance(save_to_file, str):
            return open(save_to_file, "wb")  # Shouldn't this be 'ab', for appends?
        else:
            return save_to_file  # assume it's already a stream

    def clog(*args, **kwargs):
        if verbose:
            print(*args, **kwargs)

    buffer_items = BaseBufferItems(
        input_device_index=input_device_index,
        sr=sr,
        sample_width=sample_width,
        chk_size=chk_size,
        stream_buffer_size_s=stream_buffer_size_s,
    )
    with buffer_items:
        """keep open and save to file until stop event"""
        clog("starting the recording (you can KeyboardInterrupt at any point)...")
        with get_write_file_stream() as write_stream:
            for item in buffer_items:
                try:
                    write_stream.write(item.bytes)
                except KeyboardInterrupt:
                    clog("stopping the recording...")
                    break

    clog("Done.")



import numpy as np

from taped.util import (
    DFLT_SR,
    DFLT_SAMPLE_WIDTH,
    DFLT_CHK_SIZE,
    DFLT_STREAM_BUF_SIZE_S,
    waveform_to_bytes,
)
from taped.scrap.audio_pokes import live_wf_ctx

######################################################################################################
# Example applications

from itertools import islice
import pyaudio
from time import sleep
import soundfile as sf
from io import BytesIO


def asis(wf):
    return wf


def reverse_and_print(wf):
    print("reversed sounds like this...")
    return wf[::-1]


def listen_and_shout(
    transform_wf=asis,
    every_seconds=1,
    input_device_index=None,
    sr=DFLT_SR,
    sample_width=DFLT_SAMPLE_WIDTH,
    chk_size=DFLT_CHK_SIZE,
    stream_buffer_size_s=DFLT_STREAM_BUF_SIZE_S,
):
    """

    :param transform_wf: Callable that will be called on recorded waveform before outputting to speakers
    :param every_seconds: Frequency
    :param input_device_index: Index of Input Device to use. Unspecified (or None) uses default device.
    :param sr: Specifies the desired sample rate (in Hz)
    :param sample_width: Sample width in bytes (1, 2, 3, or 4)
    :param chk_size:
    :param stream_buffer_size_s: How many seconds of data to keep in the buffer (i.e. how far in the past you can see)

    """
    # Create an interface to PortAudio
    p = pyaudio.PyAudio()

    if sample_width != 2:
        from warnings import warn

        warn("I've never seen it work with anything than sample_width=2")
    # 'output = True' indicates that the sound will be played rather than recorded
    stream = p.open(
        format=sample_width,
        channels=1,
        rate=int(sr / sample_width),  # why? I don't know. I guess unit is bytes here?
        output=True,
    )

    with live_wf_ctx(
        input_device_index,
        sr=sr,
        sample_width=sample_width,
        chk_size=chk_size,
        stream_buffer_size_s=stream_buffer_size_s,
    ) as wf_gen:
        while True:
            try:
                wf = list(islice(wf_gen, int(sr * every_seconds)))
                b = waveform_to_bytes(transform_wf(wf), sr, sample_width)
                stream.write(b)
            except KeyboardInterrupt:
                print("KeyboardInterrupt... Closing down")
                break

    # Close and terminate the stream
    stream.close()
    p.terminate()


def vol(wf):
    return np.std(np.abs(wf))


def print_vol_num(wf):
    print(f"{vol(wf):0.04f}")


def print_vol(wf, char="-", gain=2, saturation_vol=99):
    log_vol = int(min(saturation_vol, max(1, gain * np.std(np.abs(wf)) / 100)))
    print(f"{char * log_vol}")


def push_sound_through_a_pipe(
    callback=print_vol_num,
    every_seconds=1,
    input_device_index=None,
    sr=DFLT_SR,
    sample_width=DFLT_SAMPLE_WIDTH,
    chk_size=DFLT_CHK_SIZE,
    stream_buffer_size_s=DFLT_STREAM_BUF_SIZE_S,
):
    """

    :param transform_wf: Callable that will be called on recorded waveform before outputting to speakers
    :param every_seconds: Frequency
    :param input_device_index: Index of Input Device to use. Unspecified (or None) uses default device.
    :param sr: Specifies the desired sample rate (in Hz)
    :param sample_width: Sample width in bytes (1, 2, 3, or 4)
    :param chk_size:
    :param stream_buffer_size_s: How many seconds of data to keep in the buffer (i.e. how far in the past you can see)

    """
    with live_wf_ctx(
        input_device_index,
        sr=sr,
        sample_width=sample_width,
        chk_size=chk_size,
        stream_buffer_size_s=stream_buffer_size_s,
    ) as wf_gen:
        while True:
            try:
                callback(list(islice(wf_gen, int(sr * every_seconds))))
            except KeyboardInterrupt:
                print("KeyboardInterrupt... Closing down")
                break



"""
Display utilities for audio waveforms.

This module provides functions for visualizing and playing audio waveforms
in Jupyter notebooks and other Python environments. It relies on matplotlib
for visualization and IPython.display.Audio for playback.

Functions:
    getmodulename(obj, default=''): Get the name of an object's module.
    plot_wf(wf, sr=None, **kwargs): Plot a waveform against time using matplotlib.
    disp_wf(wf, sr=44100, autoplay=False, wf_plot_func=plt.specgram):
        Display a waveform visualization and return an Audio widget.
    play_wf(wf, sr=44100, autoplay=True): Play a waveform using IPython's Audio widget.

Note:
    There appears to be an incomplete import statement in the module.

"""

from inspect import getmodule
import matplotlib.pylab as plt
from numpy import linspace


def getmodulename(obj, default=""):
    """Get name of module of object"""
    return getattr(getmodule(obj), "__name__", default)


def plot_wf(wf, sr=None, **kwargs):
    if sr is not None:
        plt.plot(linspace(start=0, stop=len(wf) / float(sr), num=len(wf)), wf, **kwargs)
    else:
        plt.plot(wf, **kwargs)


def disp_wf(wf, sr=44100, autoplay=False, wf_plot_func=plt.specgram):
    if wf_plot_func is not None:
        if getmodulename(wf_plot_func, "").startswith("matplotlib"):
            plt.figure(figsize=(16, 5))
        wf_plot_func(wf)
    try:
        from IPython.display import Audio

        return Audio(data=wf, rate=sr, autoplay=autoplay)
    except:
        pass


def play_wf(wf, sr=44100, autoplay=True):
    """Play waveform using IPython Audio"""
    try:
        from IPython.display import Audio

        return Audio(data=wf, rate=sr, autoplay=autoplay)
    except:
        pass
